@@ -22,7 +22,6 @@
 
 #include <grpc/support/alloc.h>
 #include <grpc/support/log.h>
-#include <grpc/support/string_util.h>
 #include <grpc/support/time.h>
 #include <inttypes.h>
 #include <pthread.h>
@@ -30,8 +29,12 @@
 #include <stdio.h>
 #include <string.h>
 #include <time.h>
+#include <string>
 
-static intptr_t gettid(void) { return (intptr_t)pthread_self(); }
+#include "absl/strings/str_format.h"
+#include "src/core/lib/gprpp/examine_stack.h"
+
+static intptr_t sys_gettid(void) { return (intptr_t)pthread_self(); }
 
 void gpr_log(const char* file, int line, gpr_log_severity severity,
              const char* format, ...) {
@@ -83,13 +86,20 @@ void gpr_default_log(gpr_log_func_args* args) {
     strcpy(time_buffer, "error:strftime");
   }
 
-  char* prefix;
-  gpr_asprintf(&prefix, "%s%s.%09d %7" PRIdPTR " %s:%d]",
-               gpr_log_severity_string(args->severity), time_buffer,
-               (int)(now.tv_nsec), gettid(), display_file, args->line);
+  std::string prefix = absl::StrFormat(
+      "%s%s.%09d %7" PRIdPTR " %s:%d]", gpr_log_severity_string(args->severity),
+      time_buffer, (int)(now.tv_nsec), sys_gettid(), display_file, args->line);
 
-  fprintf(stderr, "%-70s %s\n", prefix, args->message);
-  gpr_free(prefix);
+  absl::optional<std::string> stack_trace =
+      args->severity >= GPR_LOG_SEVERITY_ERROR
+          ? grpc_core::GetCurrentStackTrace()
+          : absl::nullopt;
+  if (stack_trace) {
+    fprintf(stderr, "%-70s %s\n%s\n", prefix.c_str(), args->message,
+            stack_trace->c_str());
+  } else {
+    fprintf(stderr, "%-70s %s\n", prefix.c_str(), args->message);
+  }
 }
 
 #endif /* defined(GPR_POSIX_LOG) */
